/* aliM1543CPibc.c - ALi M1543C PCI-ISA Bridge Controller driver */

/* Copyright 2000-2001 Wind River Systems, Inc. */
#include "copyright_wrs.h"

/*
modification history
--------------------
01e,20mar02,zl   fixed sysPicIntDisable() return value.
01d,12sep01,h_k  changed back sysPibcInit to avoid compile error on Diab (SPR
                 #69837).
01c,11jul01,h_k  replaced value in sysPibcInit to macros. removed a work
                 around for IRQ8 asserted issue.
01b,12jun01,h_k  cleaned up sysPibcInit, aliM1543CPmuInit and aliM1543CUsbInit.
01a,09apr01,zl   written.
*/

/*
DESCRIPTION
This file contains support routines for the following:
  - the PCI-ISA bridge (PIBC) of the ALi M1543C chip
  - interrupt controller driver for the Programmable Interrupt Controller (PIC)
    of the ALi M1543C chip.
  - initialization for the SuperIO, Power Management Unit (PMU) and USB 
    host controller of the ALi M1543C chip.

SEE ALSO: Data Sheet for M1543C B1: PCI-ISA Bus Bridge with Super I/O & Fast IR
*/

/* includes */

#include "aliM1543CPibc.h"

/* PIBC access macros */

#define PIBC_BYTE_OUT(reg,data) 					\
    pciConfigOutByte (pibcPciBus, pibcPciDev, pibcPciFunc, reg, data);
#define PIBC_BYTE_IN(reg,pData) 					\
    pciConfigInByte (pibcPciBus, pibcPciDev, pibcPciFunc, reg, pData);

/* SuperIO access macros */

#define SUPERIO_BYTE_OUT(reg,data)					\
    sysOutByte ((reg) + ISA_BASE_LOCAL, (data))

#define SUPERIO_BYTE_IN(reg,pData)					\
    *(pData) = sysInByte ((reg) + ISA_BASE_LOCAL)

#define SUPERIO_REG_WRITE(reg,data)					\
    do  { 								\
	SUPERIO_BYTE_OUT (SUPERIO_INDEX_PORT, reg);			\
	SUPERIO_BYTE_OUT (SUPERIO_DATA_PORT, data);			\
	}								\
    while (0)

#define SUPERIO_REG_READ(reg,pData)					\
    do  { 								\
	SUPERIO_BYTE_OUT (SUPERIO_INDEX_PORT, reg);			\
	SUPERIO_BYTE_IN (SUPERIO_DATA_PORT, pData);			\
	}								\
    while (0)

/* RTC access macros */

#define RTC_BYTE_OUT(reg,data)						\
    sysOutByte ((reg) + ISA_BASE_LOCAL, (data))

#define RTC_BYTE_IN(reg,pData)						\
    *(pData) = sysInByte ((reg) + ISA_BASE_LOCAL)

#define RTC_REG_WRITE(reg,data)						\
    do  { 								\
	RTC_BYTE_OUT (RTC_AREG, reg);					\
	RTC_BYTE_OUT (RTC_DREG, data);					\
	}								\
    while (0)

#define RTC_REG_READ(reg,pData)						\
    do  { 								\
	RTC_BYTE_OUT (RTC_AREG, reg);					\
	RTC_BYTE_IN (RTC_DREG, pData);					\
	}								\
    while (0)


/* PIC access macros */

#define PIC_BYTE_OUT(reg,data)						\
    sysOutByte ((reg) + ISA_BASE_LOCAL, (data))

#define PIC_BYTE_IN(reg,pData)						\
    *(pData) = sysInByte ((reg) + ISA_BASE_LOCAL)

typedef struct intHandlerDesc     
    {
    VOIDFUNCPTR		rtn;		/* interrupt routine */
    int			arg;		/* interrupt handler argument */
    } INT_HANDLER_DESC;

/* forward declarations */

LOCAL STATUS	sysPicIntConnect (VOIDFUNCPTR *, VOIDFUNCPTR, int);
LOCAL void	sysPicEndOfInt (int intLevel);
LOCAL void	sysPicIntLevelSet (int intLevel);
LOCAL void	sysPicIntHandler (int arg);
LOCAL STATUS	aliM1543CSuperIoInit ();
LOCAL void	aliM1543CPmuInit (void);
LOCAL void 	aliM1543CPicInit (void);
LOCAL void 	aliM1543CUsbInit (void);

IMPORT STATUS	(*_func_intConnectHook) (VOIDFUNCPTR *vec, 
					 VOIDFUNCPTR rtn, int arg);

/* locals */

LOCAL INT_HANDLER_DESC sysPicIntTbl [PIC_INT_TABLESIZE];

/* Mask values are the currently disabled sources */

LOCAL UINT8	sysPicMask1 = 0xfb;	/* disabled all IRQs */
LOCAL UINT8	sysPicMask2 = 0xff;

/* Level values are the interrupt level masks */

LOCAL UINT8	sysPicLevel1 = 0;
LOCAL UINT8	sysPicLevel2 = 0;
LOCAL UINT8	sysPicLevelCur = 16;	/* enabled all levels */

LOCAL int	pibcPciBus = 0;
LOCAL int	pibcPciDev = 2;
LOCAL int	pibcPciFunc = 0;

/* level values by real priority */

LOCAL UCHAR sysPicPriMask1[PIC_INT_TABLESIZE+1] =
    {
    0xfb,0xfa,0xf8,0xf8,0xf0,0xe0,0xc0,0x80,
    0xf8,0xf8,0xf8,0xf8,0xf8,0xf8,0xf8,0xf8,0x0
    };

LOCAL UCHAR sysPicPriMask2 [PIC_INT_TABLESIZE+1] =
    {
    0xff,0xff,0xff,0x00,0x00,0x00,0x00,0x00,
    0xff,0xfe,0xfc,0xf8,0xf0,0xe0,0xc0,0x80,0x0
    };

/* 
 * PCI-to-ISA interrupt routing table. Value 0 means the corresponding IRQ (i.e.
 * 0, 2, 8 and  13) cannot be used, and disables the corresponding interrupt.
 */

LOCAL UCHAR sysPicIRQMap [PIC_INT_TABLESIZE] = 
    {
    0x0, 0x8, 0x0, 0x2, 0x4, 0x5, 0x7, 0x6,
    0x0, 0x1, 0x3, 0x9, 0xb, 0x0, 0xd, 0xf
    };


/*******************************************************************************
*
* sysPibcInit - Initialize the PIBC  (PCI-ISA Bridge Controller)
*
* This routine initializes the non PCI header configuration registers
* of the PIBC within the M1543C Interrupt Controller.
*
* RETURNS:  OK always
*
*/

STATUS  sysPibcInit (void)
    {
    /* PIC */
    PIBC_BYTE_OUT (CFG_REG_INDEX_PIC, PIC_DMA_LINE | PIC_ISA_MASTER_LINE | PIC_PCI_TO_ISA | PIC_DELAY_TRANSACT | PIC_PASSIVE_REL | PIC_SELECT_ISA_MASTER | PIC_SOUND);

    /* IORC - I/O Recovery control 0.25 us */
    PIBC_BYTE_OUT (CFG_REG_INDEX_IORC, IORC_ISA | IORC_ONCHIP | IORC_0_25);

    /* ISAC1 */
    PIBC_BYTE_OUT (CFG_REG_INDEX_ISAC1, ISAC1_7_16);

    /* ISAC2 */
    PIBC_BYTE_OUT (CFG_REG_INDEX_ISAC2, ISAC2_ISA_IO_NORM_16 | ISAC2_PORT_92H_SELECT);

    /* IDENR1 PCISoft reset and IDE1 interrupt selection */
    PIBC_BYTE_OUT (CFG_REG_INDEX_IDENR1, IDENR1_PRIM_CHAN_INT | IDENR1_SOFT_RESET_CTL | sysPicIRQMap[PIC_IDE1]);

    /* IDENR2 PCISoft reset and IDE2 interrupt selection */
    PIBC_BYTE_OUT (CFG_REG_INDEX_IDENR2, IDENR2_SEC_CHAN_INT_SIG | sysPicIRQMap[PIC_IDE2]);

    /* Enable USB and USB interrupt selection */
    PIBC_BYTE_OUT (CFG_REG_INDEX_USBIR, USBIR_MASTER_INTAJ | sysPicIRQMap[PIC_USB]);

    /* PIPM */
    PIBC_BYTE_OUT (CFG_REG_INDEX_PIPM, PIPM_PARITY_CHECK | PIPM_DISTR_DMA | PIPM_DELAY_TRAN_TIMEOUT);

    /* BCSC */
    PIBC_BYTE_OUT (CFG_REG_INDEX_BCSC, BCSC_ROM_64KB);

    /* PIRT1 - Set INTA and INTB IRQ numbers */
    PIBC_BYTE_OUT (CFG_REG_INDEX_PIRT1, sysPicIRQMap[PIC_INTB] << 4 | sysPicIRQMap[PIC_INTA]);

    /* PIRT2 - INTC and INTD IRQ numbers */
    PIBC_BYTE_OUT (CFG_REG_INDEX_PIRT2, sysPicIRQMap[PIC_INTD] << 4 | sysPicIRQMap[PIC_INTC]);

    /* PIRT3 */
    PIBC_BYTE_OUT (CFG_REG_INDEX_PIRT3, PIRT_INT_DISABLE);

    /* PIRT4 */
    PIBC_BYTE_OUT (CFG_REG_INDEX_PIRT4, PIRT_INT_DISABLE);

    /* PILET - Level to edge trigger transfer for INTA-INTD */
    PIBC_BYTE_OUT (CFG_REG_INDEX_PILET, PILET_INT1J | PILET_INT2J | PILET_INT3J | PILET_INT4J);

    /* IDEIC - Enable IDE controller */
    PIBC_BYTE_OUT (CFG_REG_INDEX_IDEIC, IDEIC_SIRQI_SIRQII | IDEIC_ATA_PRI_SIG_PAD | IDEIC_ATA_SEC_SIG_PAD | IDEIC_IDSEL_A27 | IDEIC_ONCHIP_IDE_CONTR);

    /* Initialize the Power Management Unit */
    aliM1543CPmuInit();

    /* Initialize USB device */
    aliM1543CUsbInit();

    /* Initialize the SuperIO */
    aliM1543CSuperIoInit ();

    /* Initialize the Interrupt Controller */
    aliM1543CPicInit();

    return (OK);
    }


/*******************************************************************************
*
* sysPibcHwInit2 - connect the PIC interrupt handler
*
* This routine connects the PIC interrupt handler. The PIC interrupt handler
* is responsabile for determining and invoking the actual user routine.
*
* NOMANUAL
*
* RETURNS: OK.
*/

LOCAL STATUS  sysPibcHwInit2 ()
    {
    intConnect (INT_VEC_PIBC, sysPicIntHandler, 0);

    /* install the BSP specific interrupt connect routine */

    _func_intConnectHook =  sysPicIntConnect;

    return (OK);
    }


/*******************************************************************************
*
* sysPicIntEnable - enable a PIC interrupt level
*
* This routine enables a specified PIC interrupt level.
*
* NOMANUAL
*
* RETURNS: OK or ERROR.
*/

STATUS sysPicIntEnable
    (
    int intNum
    )
    {
    if (intNum < 16)
        {
        if (intNum < 8)
            {
            sysPicMask1 &= ~(1 << intNum);
            PIC_BYTE_OUT (PIC1_OCW1, sysPicMask1 | sysPicLevel1);
            }
        else
            {
            sysPicMask2 &= ~(1 << (intNum - 8));
            PIC_BYTE_OUT (PIC2_OCW1, sysPicMask2 | sysPicLevel2);
            }
        return (OK);
        }
    else
        {
        return ERROR;
        }
    }

 
/*******************************************************************************
*
* sysPicIntDisable - disable a PIC interrupt level
*
* This routine disables a specified PIC interrupt level.
*
* NOMANUAL
*
* RETURNS: OK, or ERROR when intNum is out of valid range.
*/

STATUS sysPicIntDisable
    (
    int intNum	/*  interrupt level to disable  */
    )
    {

    if ((intNum < 16) && (intNum >= 0))
        {
        if (intNum < 8)
            {
            sysPicMask1 |= (1 << intNum);
            PIC_BYTE_OUT (PIC1_OCW1, sysPicMask1 | sysPicLevel1 );
            }
        else
            {
            sysPicMask2 |= (1 << (intNum - 8));
            PIC_BYTE_OUT (PIC2_OCW1, sysPicMask2 | sysPicLevel2);
            }
        return (OK);
        }
    else
        {
        return (ERROR);
        }
    }


/*******************************************************************************
*
* sysPicIntHandler - handle a PIC interrupt
*
* This routine handles an interrupt which may have originated from the
* PIBC chip's 8259 controllers.
*
* NOMANUAL
*
* RETURNS: N/A
*/

LOCAL void sysPicIntHandler
    (
    int arg
    )
    {
    int		oldLevel;
    UCHAR 	irrVal1;
    UCHAR 	irrVal2;
    UCHAR 	picIrq;


    /* read the IRR-s and check if the interrupt came from the PIBC. */
    
    PIC_BYTE_OUT (PIC1_OCW3, 0x08+0x02);
    PIC_BYTE_IN (PIC1_OCW3, &irrVal1);

    PIC_BYTE_OUT (PIC2_OCW3, 0x08+0x02);
    PIC_BYTE_IN (PIC2_OCW3, &irrVal2);

    /* If there are no pending PIC interrupts, return */

    if ((irrVal1 | irrVal2) == 0)
        {
        return;
        }

    /* 
     * Get the highest priority IRQ pending. The priority order is:
     * 0, 1, 8, 9, 10, 11, 12, 13, 14, 15, 3, 4, 5, 6, 7
     */

    picIrq = 0;
    while ((irrVal1 & 1) == 0)
	{
	irrVal1 = irrVal1 >> 1;
	picIrq++;
	}

    if (picIrq == 2)				/* check if cascaded */
	{
	picIrq = 8;
	while ((irrVal2 & 1) == 0)
	    {
	    irrVal2 = irrVal2 >> 1;
	    picIrq++;
	    }
	}

    /* mask this interrupt level and lower priority interrupt levels */

    oldLevel = sysPicLevelCur;

    sysPicIntLevelSet (picIrq); 

    /* Re-arm (enable) the interrupt chip */

    sysPicEndOfInt (picIrq);

    /*  call the interrupt handler */

    if (sysPicIntTbl[picIrq].rtn == NULL)
        {
	logMsg ("uninitalized PIC interrupt %x\r\n",
                picIrq, 0,0,0,0,0);
        }
    else
        {
  	(sysPicIntTbl[picIrq].rtn) (sysPicIntTbl[picIrq].arg);
        }
        
    /*  set the original interrupt mask back to original value */

    sysPicIntLevelSet (oldLevel);
    }


/*******************************************************************************
*
* sysPicEndOfInt - send EOI(end of interrupt) signal.
*
* This routine is called at the end of the interrupt handler to
* send a non-specific end of interrupt (EOI) signal.
*
* The second PIC is acked only if the interrupt came from that PIC.
* The first PIC is always acked.
*
* RETURNS: N/A
*/

LOCAL void  sysPicEndOfInt
    (
    int intNum
    )
    {
    if (intNum > 7)
	{
        PIC_BYTE_OUT (PIC2_OCW2, 0x20);
	}
    PIC_BYTE_OUT (PIC1_OCW2, 0x20);
    }


/*******************************************************************************
*
* sysPicIntLevelSet - set the interrupt priority level
*
* NOTES
*
* This routine masks interrupts with real priority equal to or lower than
* <intNum>.  The special
* value 16 indicates all interrupts are enabled. Individual interrupt
* numbers have to be specifically enabled by sysPicIntEnable() before they
* are ever enabled by setting the interrupt level value.
*
* Note because of the IBM cascade scheme, the actual priority order for
* interrupt numbers is (high to low) 0, 1, 8, 9, 10, 11, 12, 13, 14, 15,
* 3, 4, 5, 6, 7, 16 (all enabled)
*
* INTERNAL: It is possible that we need to determine if we are raising
* or lowering our priority level.  It may be that the order of loading the
* two mask registers is dependent upon raising or lowering the priority.
*
* RETURNS: N/A
*/

LOCAL void  sysPicIntLevelSet
    (
    int intNum   /*  interrupt level to implement  */
    )
    {
    if (intNum < 0)
	return;		/* invalid value */

    if (intNum > 16)
	intNum = 16;

    sysPicLevelCur = intNum;

    if (sysPicLevel2 != sysPicPriMask2[intNum])
	{
	sysPicLevel2 = sysPicPriMask2[intNum];
	PIC_BYTE_OUT (PIC2_OCW1, sysPicMask2 | sysPicLevel2);
	}

    if (sysPicLevel1 != sysPicPriMask1[intNum])
	{
	sysPicLevel1 = sysPicPriMask1[intNum];
	PIC_BYTE_OUT (PIC1_OCW1, sysPicMask1 | sysPicLevel1);
	}
    }


/*******************************************************************************
*
* sysPicIntConnect - Connect a user routine to the PIC interrupt handler
*
* This routine connects a user routine to the PIC interrupt handler if the
*
* RETURNS:  OK always
*/

LOCAL STATUS sysPicIntConnect
    (
    VOIDFUNCPTR *vec,
    VOIDFUNCPTR rtn,
    int		arg
    )
    {
    STATUS	status = ERROR;

    if (((UINT)vec >= (UINT) PIC_IVEC_BASE) && 
	((UINT)vec <  (UINT) PIC_IVEC_BASE + 4 * PIC_INT_TABLESIZE))
	{
	UCHAR	picIrq = PIC_IVEC_TO_IRQ (vec);

	/* Connect to the PIC vector table */

	sysPicIntTbl[picIrq].rtn = rtn;
	sysPicIntTbl[picIrq].arg = arg;

	/* Only for host tools (browser) */

	intVecSet ((FUNCPTR *) vec, (FUNCPTR) rtn);

	status = OK;
	}
    else
	{
	/* Connect to the CPU vector table */
	
	FUNCPTR intDrvRtn = intHandlerCreate ((FUNCPTR)rtn, arg);

	if (intDrvRtn != NULL)
	    {
	    /* make vector point to synthesized code */

	    intVecSet ((FUNCPTR *)vec, (FUNCPTR)intDrvRtn);

	    status = OK;
	    }
	}
    return status;
    }


/*******************************************************************************
*
*  aliM1543CPicInit - Initialize the two 8259 PICs of the M1543C
*
* This routine initializes the two Programmable Interrupt Controolers
* of the M1543C PCI-ISA Bridge.
*
* RETURNS:  OK always
*/

LOCAL void aliM1543CPicInit (void)
    {
    /*  Initialize the Interrupt Controller #1 as master  */

    PIC_BYTE_OUT (PIC1_ICW1, 0x11);
    PIC_BYTE_OUT (PIC1_ICW2, 0);
    PIC_BYTE_OUT (PIC1_ICW3, 0x04);
    PIC_BYTE_OUT (PIC1_ICW4, 0x01);

    /*  Initialize the Interrupt Controller #2 as slave */

    PIC_BYTE_OUT (PIC2_ICW1, 0x11);
    PIC_BYTE_OUT (PIC2_ICW2, 8);
    PIC_BYTE_OUT (PIC2_ICW3, 0x02);
    PIC_BYTE_OUT (PIC2_ICW4, 0x01);

    /* Mask interrupts (except cascaded one) by writing to OCW1 register */

    PIC_BYTE_OUT (PIC1_OCW1, 0xfb);
    PIC_BYTE_OUT (PIC2_OCW1, 0xff);
 
    /* Set level to 16 */
    sysPicIntLevelSet (16);

    /* Enable cascaded interrupts */
    sysPicIntEnable (2);

    /*  perform the end of interrupt procedure  */
    sysPicEndOfInt (15); 
    }


/*******************************************************************************
*
* aliM1543CSuperIOInit - Initialize the SuperIO on M1543C
*
* This routine initializes the SuperIO devices on the M1543C.
*
* RETURNS:  OK or ERROR
*/

LOCAL STATUS aliM1543CSuperIoInit ()
    {
    UCHAR	idHigh;
    UCHAR	idLow;
    /* Enter configuration state */
    
    SUPERIO_BYTE_OUT (SUPERIO_CFG_PORT, CFG_KEY1);
    SUPERIO_BYTE_OUT (SUPERIO_CFG_PORT, CFG_KEY2);

    /* Verify device ID */
    
    SUPERIO_REG_READ (INDEX_IDH, &idHigh);
    SUPERIO_REG_READ (INDEX_IDL, &idLow);

    if (((idHigh << 8) | idLow) != M1543SIO_ID)
	{
	return (ERROR);
	}

    /* Reset chip */

    SUPERIO_REG_WRITE (INDEX_RESET, 0x01);

    /* Configure Logical device 4 - UART1 */

    SUPERIO_REG_WRITE (INDEX_DEV, LOG_DEV_UART1);
    SUPERIO_REG_WRITE (INDEX_DEVEN, DEV_EN);
    SUPERIO_REG_WRITE (INDEX_IOBASEH, MSB(UART1_OFFSET));
    SUPERIO_REG_WRITE (INDEX_IOBASEL, LSB(UART1_OFFSET));
    SUPERIO_REG_WRITE (INDEX_IRQSEL1, PIC_UART1);

    /* Configure Logical device 5 - UART3 */

    SUPERIO_REG_WRITE (INDEX_DEV, LOG_DEV_UART2);
    SUPERIO_REG_WRITE (INDEX_DEVEN, DEV_EN);
    SUPERIO_REG_WRITE (INDEX_IOBASEH, MSB(UART2_OFFSET));
    SUPERIO_REG_WRITE (INDEX_IOBASEL, LSB(UART2_OFFSET));
    SUPERIO_REG_WRITE (INDEX_IRQSEL1, PIC_UART2);

    /* Configure Logical device 7 - keyboard and mouse */

    SUPERIO_REG_WRITE (INDEX_DEV, LOG_DEV_KBD);
    SUPERIO_REG_WRITE (INDEX_DEVEN, DEV_EN);
    SUPERIO_REG_WRITE (INDEX_IRQSEL1, PIC_KBD);
    SUPERIO_REG_WRITE (INDEX_IRQSEL2, PIC_MOUSE);

    SUPERIO_BYTE_OUT (SUPERIO_CFG_PORT, EXIT_KEY);

    return (OK);
    }


/*******************************************************************************
*
* aliM1543CPmuInit - Initialize the Power Management Unit on M1543C
*
* This routine initializes the Power Management Unit on the M1543C.
*
* RETURNS:  N/A
*/

LOCAL void aliM1543CPmuInit ()
    {
    int		pciBusNo;
    int		pciDevNo;
    int		pciFuncNo;

    /* PMU base address registers must be within the first 64k */
    
    if (pciFindDevice(PMU_VENDOR_ID, PMU_DEVICE_ID, 0,
                      &pciBusNo, &pciDevNo, &pciFuncNo) != ERROR)
        {
        pciConfigOutLong (pciBusNo, pciDevNo, pciFuncNo,
                      PCI_CFG_BASE_ADDRESS_0, PMU_ABA_IO_BASE_ADRS);
        pciConfigOutLong (pciBusNo, pciDevNo, pciFuncNo,
                      PCI_CFG_BASE_ADDRESS_1, PMU_SBA_IO_BASE_ADRS);
        }
    }


/*******************************************************************************
*
* aliM1543CUsbInit - Initialize the USB on M1543C
*
* This routine initializes the USB device on the M1543C.
*
* RETURNS:  N/A
*/

LOCAL void aliM1543CUsbInit ()
    {
    int		pciBusNo;
    int		pciDevNo;
    int		pciFuncNo;

    if (pciFindDevice(USB_VENDOR_ID, USB_DEVICE_ID, 0,
                      &pciBusNo, &pciDevNo, &pciFuncNo) != ERROR)
        {
        pciConfigOutWord (pciBusNo, pciDevNo, pciFuncNo, PCI_CFG_COMMAND, 
		      PCI_CMD_MEM_ENABLE | PCI_CMD_MASTER_ENABLE);
        }
    }


/*******************************************************************************
*
* sysRTCRead - Read from the RTC interface
*
* This routine reads from the RTC interface
*
* RETURNS:  N/A
*/

UCHAR sysRTCRead
    (
    UCHAR reg
    )
    {
    UCHAR val;
    RTC_REG_READ (reg, &val);
    return (val);
    }

/*******************************************************************************
*
* sysRTCWrite - Write to the RTC interface
*
* This routine writes to the RTC interface
*
* RETURNS:  N/A
*/

void sysRTCWrite
    (
    UCHAR reg,
    UCHAR val
    )
    {
    RTC_REG_WRITE (reg, val);
    }
