/* sysAta.c - Hitachi MS7751SE01 series ATA-2 initialization for sysLib.c */

/* Copyright 1984-2002 Wind River Systems, Inc. */
#include "copyright_wrs.h"

/*
modification history
--------------------
01c,01may02,zl   added sysAtaConfigure() for timing regs optimizations.
01b,29apr02,zl   fixed IO routines (SPR#75073); timing regs initialization.
01a,23apr00,zl   written.
*/

/* 
DESCRIPTION

This file contains the sysAtaInit() necessary for
initializing the ATA/EIDE subsystem. 
*/

/* includes */

#include "vxWorks.h"
#include "config.h"

#define IDE_VENDOR_ID		0x10b9		/* IDE on M1543C */
#define IDE_DEVICE_ID		0x5229

#define PCI_CFG_IDE_PCAS	0x58		/* primary ctrl timing regs. */
#define PCI_CFG_IDE_PCCB	0x59
#define PCI_CFG_IDE_PCDT0	0x5a
#define PCI_CFG_IDE_PCDT1	0x5b
#define PCI_CFG_IDE_SCAS	0x5c		/* secondary ctrl timing regs */
#define PCI_CFG_IDE_SCCB	0x5d
#define PCI_CFG_IDE_SCDT0	0x5e
#define PCI_CFG_IDE_SCDT1	0x5f

/* typedefs */

typedef struct ideTimingsAli
    {
    UINT8 setup;				/* adrs. setup cycle count */
    UINT8 actRec;				/* active and recovery cycle */
						/* count in form: */
						/*   actRec = act << 4 | rec */
    } IDE_TIMINGS_ALI;

/* imports */

IMPORT VOIDFUNCPTR _func_sysAtaInit;		/* from ataDrv.c */
IMPORT ATA_CTRL   ataCtrl [];			/* from ataDrv.c */

/* forward declarations */

LOCAL void sysAtaConfigure (int ctrl);

/* global data structures used by the EIDE/ATA driver */
 
ATA_RESOURCE ataResources[ATA_MAX_CTRLS];
 
/*
 * The first member in the ATA_TYPE struct has a dual purpose.
 *   1) If cylinders == 0, the device location is not probed at startup.
 *   2) If cylinders |= 0, the device location is probed, and if a device
 *      is found, the driver will fill in the first 3 member of the struct
 *      with number of cylinders, number of heads, and sectors per track.
 *
 * The last 2 members of the struct are static and should not be changed.
 *
 * The ATA_TYPE struct has the following layout:
 *      int cylinders;
 *      int heads;
 *      int sectorsTrack;
 *      int bytesSector;
 *      int precomp;
 *
 * NOTE: If configType == ATA_GEO_FORCE, the user needs to fill in
 * values for cylinders, heads, and sectorsTrack. Setting cylinders to
 * zero disables that drive.
 */
 
ATA_TYPE ataTypes [ATA_MAX_CTRLS][ATA_MAX_DRIVES] =
    {
        {
            { 1, 0, 0, 512, 0xff},  /* controller 0, drive 0 */
            { 0, 0, 0, 512, 0xff},  /* controller 0, drive 1 */
        },
        {
            { 1, 0, 0, 512, 0xff},  /* controller 1, drive 0 */
            { 0, 0, 0, 512, 0xff},  /* controller 1, drive 1 */
        }
    };

/*
 * The following table represents the PCI cycle counts at 33MHz for the
 * IDE timing specification. The first values represents address setup cycle
 * count. The second value is the data active and recovery cycle count in
 * form of: actRec = act << 4 | rec.
 * For details see the M1543C Data Sheet and the ATA specification.
 */

LOCAL IDE_TIMINGS_ALI ideTimings[] = 
    {
			/*              setup  active  rec.  Cycle */
			/*              =====  ======  ====  ===== */
	{0x03, 0x6c},   /* PIO 0 timings:  70     165   365    600 */
	{0x02, 0x57},   /* PIO 1 timings:  50     125   208    383 */
	{0x01, 0x44},   /* PIO 2 timings:  30     100   110    240 */
	{0x01, 0x33},   /* PIO 3 timings:  30      80    70    180 */
	{0x01, 0x31},   /* PIO 4 timings:  25      70    25    120 */
    };


/******************************************************************************
*
* sysAtaInit - initialize the EIDE/ATA interface
*
* Perform the necessary initialization required before starting up the
* ATA/EIDE driver.
*/

STATUS sysAtaInit ()
    {
    int                 pciBus;
    int                 pciDevice;
    int                 pciFunc;
    UINT32              iobase1Csr;
    UINT32              iobase2Csr;

    /* initialize IDE1 */

    if (pciFindDevice (IDE_VENDOR_ID, IDE_DEVICE_ID, 0,
                       &pciBus, &pciDevice, &pciFunc) != OK)
	return (ERROR);

    pciConfigInLong (pciBus, pciDevice, pciFunc,
                     PCI_CFG_BASE_ADDRESS_0, &iobase1Csr);
    pciConfigInLong (pciBus, pciDevice, pciFunc,
                     PCI_CFG_BASE_ADDRESS_1, &iobase2Csr);

    ataResources[0].resource.ioStart[0] = ISA_BASE_LOCAL | 
 					  (iobase1Csr & PCI_IOBASE_MASK);
    ataResources[0].resource.ioStart[1] = ISA_BASE_LOCAL | 
 					  ((iobase2Csr & PCI_IOBASE_MASK) + 2);

    ataResources[0].ctrlType = IDE_LOCAL;
    ataResources[0].drives = 2;
    ataResources[0].intVector = (int) IVEC_TO_INUM (PIC_IRQ_TO_IVEC(PIC_IDE1));
    ataResources[0].intLevel = (int) PIC_IDE1;
    ataResources[0].configType = ATA_PIO_AUTO     | ATA_PIO_MULTI |
				 ATA_GEO_PHYSICAL | ATA_BITS_16;
    ataResources[0].semTimeout = ATA_SEM_TIMEOUT;
    ataResources[0].wdgTimeout = ATA_WDG_TIMEOUT;

    /* 
     * Configure the primary IDE channel Timing Registers.
     * At this time use a values that are safe for the slowest PIO mode 0,
     * but small enough to avoid the SH7751 PCIC timeout interrupts.
     * These values can be optimized once the disk type has been detected
     * and the supported PIO mode has been obtained.
     */
    
    pciConfigOutByte (pciBus, pciDevice, pciFunc, PCI_CFG_IDE_PCAS, 
    		      ideTimings[0].setup);
    pciConfigOutByte (pciBus, pciDevice, pciFunc, PCI_CFG_IDE_PCCB,
		      ideTimings[0].actRec);
    pciConfigOutByte (pciBus, pciDevice, pciFunc, PCI_CFG_IDE_PCDT0,
		      ideTimings[0].actRec);
    pciConfigOutByte (pciBus, pciDevice, pciFunc, PCI_CFG_IDE_PCDT1,
		      ideTimings[0].actRec);

    /* Second controller */

    pciConfigInLong (pciBus, pciDevice, pciFunc,
                     PCI_CFG_BASE_ADDRESS_2, &iobase1Csr);
    pciConfigInLong (pciBus, pciDevice, pciFunc,
                     PCI_CFG_BASE_ADDRESS_3, &iobase2Csr);

    ataResources[1].resource.ioStart[0] = ISA_BASE_LOCAL | 
 					  (iobase1Csr & PCI_IOBASE_MASK);
    ataResources[1].resource.ioStart[1] = ISA_BASE_LOCAL | 
 					  ((iobase2Csr & PCI_IOBASE_MASK) + 2);

    ataResources[1].ctrlType = IDE_LOCAL;
    ataResources[1].drives = 2;
    ataResources[1].intVector = (int) IVEC_TO_INUM (PIC_IRQ_TO_IVEC(PIC_IDE2));
    ataResources[1].intLevel = (int) PIC_IDE2;
    ataResources[1].configType = ATA_PIO_AUTO     | ATA_PIO_MULTI |
				 ATA_GEO_PHYSICAL | ATA_BITS_16;
    ataResources[1].semTimeout = ATA_SEM_TIMEOUT;
    ataResources[1].wdgTimeout = ATA_WDG_TIMEOUT;

    /* 
     * Configure the primary IDE channel Timing Registers.
     * At this time use a values that are safe for the slowest PIO mode 0,
     * but small enough to avoid the SH7751 PCIC timeout interrupts.
     * These values can be optimized once the disk type has been detected
     * and the supported PIO mode has been obtained.
     */

    pciConfigOutByte (pciBus, pciDevice, pciFunc, PCI_CFG_IDE_SCAS, 
    		      ideTimings[0].setup);
    pciConfigOutByte (pciBus, pciDevice, pciFunc, PCI_CFG_IDE_SCCB,
		      ideTimings[0].actRec);
    pciConfigOutByte (pciBus, pciDevice, pciFunc, PCI_CFG_IDE_SCDT0,
		      ideTimings[0].actRec);
    pciConfigOutByte (pciBus, pciDevice, pciFunc, PCI_CFG_IDE_SCDT1,
		      ideTimings[0].actRec);

    _func_sysAtaInit = (VOIDFUNCPTR) sysAtaConfigure;

    return (OK);
    }


/*******************************************************************************
*
* sysAtaConfigure - finish configuration of the controller
*
* This function is called to configure the ATA controller once the hard disk
* types have been detected. It's main role is to optimize the timings based on
* the PIO modes selected.
*
* RETURNS: OK, always.
*/

LOCAL void sysAtaConfigure 
    (
    int		ctrl		/* ATA controller */
    )
    {
    int         pciBus;		/* PCI bus number */
    int		pciDevice;	/* PCI device number */
    int		pciFunc;	/* PCI function */
    ATA_CTRL * 	pCtrl;		/* ATA controller */
    ATA_DRIVE * pDrive;		/* Disk drive */
    int		i;
    int		pioMode[ATA_MAX_DRIVES];
    int		pioModeMin = 4;
    BOOL	diskFound = FALSE;

    /* check for valid parameter */

    if ((ctrl < 0) || (ctrl >= ATA_MAX_CTRLS))
	return;

    pCtrl  = &ataCtrl[ctrl];  

    if (pCtrl->installed != TRUE)
        return;

    /* find device on PCI bus */

    if (pciFindDevice (IDE_VENDOR_ID, IDE_DEVICE_ID, 0,
                       &pciBus, &pciDevice, &pciFunc) != OK)
	return;
    
    /* get PIO mode of drives */

    for (i = 0; i < ATA_MAX_DRIVES; i++)
	{
	pioMode[i] = 0;
	pDrive = &pCtrl->drive[i];

	/* check if drive was found */

	if (pDrive->type == ATA_TYPE_ATA)
	    diskFound = TRUE;
	else
	    continue;

	/* save PIO mode of drive */

	pioMode[i] = pDrive->pioMode;

	/* find worst case */

	if (pioMode[i] < pioModeMin)
	    pioModeMin = pioMode[i];
	}

    /* If there were no disks found on this controller, simply return */

    if (diskFound == FALSE)
	return;

    /* set the timing registers based on PIO modes */

    if (ctrl == 0)
	{
	/* configure primary controller */

	pciConfigOutByte (pciBus, pciDevice, pciFunc, PCI_CFG_IDE_PCAS, 
			  ideTimings[pioModeMin].setup);
	pciConfigOutByte (pciBus, pciDevice, pciFunc, PCI_CFG_IDE_PCCB,
			  ideTimings[pioModeMin].actRec);
	pciConfigOutByte (pciBus, pciDevice, pciFunc, PCI_CFG_IDE_PCDT0,
			  ideTimings[pioMode[0]].actRec);
	pciConfigOutByte (pciBus, pciDevice, pciFunc, PCI_CFG_IDE_PCDT1, 
			  ideTimings[pioMode[1]].actRec);
	}
    else
	{    
	/* configure secondary controller */

	pciConfigOutByte (pciBus, pciDevice, pciFunc, PCI_CFG_IDE_SCAS,  
			  ideTimings[pioModeMin].setup);
	pciConfigOutByte (pciBus, pciDevice, pciFunc, PCI_CFG_IDE_SCCB,
			  ideTimings[pioModeMin].actRec);
	pciConfigOutByte (pciBus, pciDevice, pciFunc, PCI_CFG_IDE_SCDT0,
			  ideTimings[pioMode[0]].actRec);
	pciConfigOutByte (pciBus, pciDevice, pciFunc, PCI_CFG_IDE_SCDT1, 
			  ideTimings[pioMode[1]].actRec);
	}
    }

/*******************************************************************************
*
* sysIntEnablePIC - enable interrupts
*
* This function enables IDE controller interrupts.
*
* RETURNS: N/A
*/
 
STATUS sysIntEnablePIC
    (
    int intNum
    )
    {
    /* Pass tyhe IRQ number to the PIC */
    
    sysPicIntEnable (intNum);

    return OK;
    }


/*****************************************************************************
*
* sysInWordString - reads a string of words from an io address.
*
* This function reads a word string from a specified io address.
*
* RETURNS: N/A
*/
 
void sysInWordString
    (
    ULONG	ioAddr,
    UINT16 *	pData,
    int		nWords
    )
    {
    int 	loopCtr;
    UINT16 	w;
    UINT8 *	pByte;

    pByte = (UINT8 *) pData;

    for (loopCtr = 0; loopCtr < nWords; loopCtr++)
	{
	w = *(volatile UINT16 *)ioAddr;
#if (_BYTE_ORDER == _BIG_ENDIAN)
	*pByte++ = MSB(w);
	*pByte++ = LSB(w);
#else
	*pByte++ = LSB(w);
	*pByte++ = MSB(w);
#endif
	}
    }
 
/*****************************************************************************
*
* sysInWordStringRev - reads a string of words that are byte reversed
*
* This function reads a string of words that are byte reversed from a
* specified io address. It is only used in big endian mode.
*
* RETURNS: N/A
*/
 
void sysInWordStringRev
    (
    ULONG	ioAddr,
    UINT16 *	pData,
    int		nWords
    )
    {
    int 	loopCtr;
    UINT16 	w;
    UINT8 *	pByte;

    pByte = (UINT8 *) pData;

    for (loopCtr = 0; loopCtr < nWords; loopCtr++)
        {
	w = *(volatile UINT16 *)ioAddr;
	*pByte++ = LSB(w);
	*pByte++ = MSB(w);
	}
    }
 
/*****************************************************************************
*
* sysOutWordString - writes a string of words to an IO address.
*
* This function writes a word string from a specified IO address.
*
* RETURNS: N/A
*/
 
void sysOutWordString
    (
    ULONG       ioAddr,
    UINT16 *    pData,
    int         nWords
    )
    {
    int 	loopCtr;
    UINT16 	w;
    UINT8 *	pByte;
    
    pByte = (UINT8 *) pData;

    for (loopCtr = 0; loopCtr < nWords; loopCtr++)
	{
	w = *pByte++;
#if (_BYTE_ORDER == _BIG_ENDIAN)
	w = (w << 8) | (*pByte++);
#else
	w = (w ) | ((*pByte++) << 8);
#endif
        *(volatile UINT16 *)ioAddr = w;
	}
    }
 
/*****************************************************************************
*
* sysInLongString - reads a string of longwords from an io address.
*
* This function reads a longword string from a specified io address.
*
* RETURNS: N/A
*/
 
void sysInLongString
    (
    ULONG	ioAddr,
    ULONG *	bufPtr,
    int		nLongs
    )
    {
    /* 
     * 32-bit ATA access is not supported on this board, but this function
     * is required for linking with ataDrv.
     */
    }
 
/*****************************************************************************
*
* sysOutLongString - writes a string of longwords to an io address.
*
* This function writes a longword string from a specified io address.
*
* RETURNS: N/A
*/
 
void sysOutLongString
    (
    ULONG       ioAddr,
    ULONG *     bufPtr,
    int         nLongs
    )
    {
    /* 
     * 32-bit ATA access is not supported on this board, but this function
     * is required for linking with ataDrv.
     */
    }
