/* sysBusPci.c - MS7751SE01 platform-specific PCI bus support */

/* Copyright 1999-2003 Wind River Systems, Inc. */
#include "copyright_wrs.h"

/*
modification history
--------------------
01g,16jun03,h_k  fixed interrupt mask clear register access. (SPR #86882)
01f,30apr02,frf  fixed enabling PCIC interrupts (SPR #76415).
01e,27nov01,zl   componentized virtual PCI support.
01d,01nov01,zl   corrected NULL usage.
01d,24oct01,hk   added virtual PCI memory and I/O support.
01c,11jun01,h_k  fixed intLine table for USB.
01b,30mar01,zl   re-written.
01a,15dec00,rsh  written for ms7751se01.
*/

/*
DESCRIPTION
This is the MS7751SE01 platform specific pciAutoConfigLib information.
*/


/* includes */

#include "vxWorks.h"
#include "config.h"
#include "sysLib.h"
#ifdef INCLUDE_SH7751_VIRTUAL_PCI_MEMIO
#include "excLib.h"
#endif /* INCLUDE_SH7751_VIRTUAL_PCI_MEMIO */

#include "drv/pci/pciConfigLib.h" 	
#include "drv/pci/pciIntLib.h" 	
#include "drv/pci/pciAutoConfigLib.h"	
#include "sysBusPci.h"

/* locals */

LOCAL PCI_SYSTEM sysParams;


/* 
 * All PCI interrupts are connected to one of the Programmable Interrupt
 * Controller (PIC) IRQs of the M1543C. The intLine Array stores the IRQ 
 * numbers. The IDE device driver uses intConnect() instead of pciIntConnect(),
 * and it uses it's own vector numbers for IDE1 and IDE2.
 */

static UCHAR intLine [][4] =
    {
    /*    IntA      IntB      IntC      IntD   */
        { PIC_INTA, PIC_INTB, PIC_INTC, PIC_INTD }, /* dev.  0 - Enet 79C973 */
        { PIC_INTB, PIC_INTC, PIC_INTD, PIC_INTA }, /* dev.  1 - PCI Slot 1  */
        { 0xff,     0xff,     0xff,     0xff     }, /* dev.  2 - PIBC M1543C */
        { PIC_INTC, PIC_INTD, PIC_INTA, PIC_INTB }, /* dev.  3 - PCI Slot 2  */
        { 0xff,     0xff,     0xff,     0xff     }, /* dev.  4 - N/A         */
        { 0xff,     0xff,     0xff,     0xff     }, /* dev.  5 - N/A         */
        { 0xff,     0xff,     0xff,     0xff     }, /* dev.  6 - N/A         */
        { 0xff,     0xff,     0xff,     0xff     }, /* dev.  7 - N/A         */
        { 0xff,     0xff,     0xff,     0xff     }, /* dev.  8 - N/A         */
        { 0xff,     0xff,     0xff,     0xff     }, /* dev.  9 - N/A         */
        { 0xff,     0xff,     0xff,     0xff     }, /* dev. 10 - N/A         */
        { 0xff,     0xff,     0xff,     0xff     }, /* dev. 11 - IDE M1543C  */
        { 0xff,     0xff,     0xff,     0xff     }, /* dev. 12 - PWR M1543C  */
        { 0xff,     0xff,     0xff,     0xff     }, /* dev. 13 - N/A         */
        { 0xff,     0xff,     0xff,     0xff     }, /* dev. 14 - N/A         */
        { PIC_USB,  PIC_USB,  PIC_USB,  PIC_USB  }  /* dev. 15 - USB M1543C  */
    };

/* Forward declarations */

void pcicInt (void);


/*******************************************************************************
*
* sysPciAutoConfig - PCI autoconfig support routine
*
* RETURNS: N/A
*/

void sysPciAutoConfig (void)
    {

    /* 32-bit Prefetchable Memory Space */

    sysParams.pciMem32 		= PCI_MEM_ADRS;
    sysParams.pciMem32Size 	= PCI_MEM_SIZE;

    /* 32-bit Non-prefetchable Memory Space */

    sysParams.pciMemIo32 	= PCI_MEMIO_ADRS;
    sysParams.pciMemIo32Size 	= PCI_MEMIO_SIZE;

    /* 16-bit ISA I/O Space */

    sysParams.pciIo16 		= PCI_ISA_IO_ADRS;
    sysParams.pciIo16Size 	= PCI_ISA_IO_SIZE;

    /* 32-bit PCI I/O Space */

    sysParams.pciIo32 		= PCI_IO_ADRS;
    sysParams.pciIo32Size 	= PCI_IO_SIZE;

    /* Configuration space parameters */

    sysParams.maxBus 		= 0;
    sysParams.cacheSize 	= _CACHE_ALIGN_SIZE / 4;
    sysParams.maxLatency 	= PCI_LAT_TIMER;
    sysParams.autoIntRouting 	= TRUE;

    /* Device inclusion and interrupt routing routines */

    sysParams.includeRtn 	= sysPciAutoconfigInclude;
    sysParams.intAssignRtn 	= sysPciAutoconfigIntrAssign;

    /* PCI-to-PCI Bridge Pre- and Post-enumeration init routines */

    sysParams.bridgePreConfigInit = NULL;
    sysParams.bridgePostConfigInit = NULL;

    /*
     * Perform any needed PCI Host Bridge
     * Initialization that needs to be done
     * before pciAutoConfig is invoked here 
     * utilizing the information in the 
     * newly-populated sysParams structure. 
     */

    sysPcicInit();
    
    /* 
     * Initialize PCI-ISA bridge. This must be after PCIC initialization and
     * before pciAutoConfig().
     */
    
    sysPibcInit();

    /* now do auto configuration */

    pciAutoConfig (&sysParams);

    /*
     * Perform any needed post-enumeration
     * PCI Host Bridge Initialization here
     * utilizing the information in the 
     * sysParams structure that has been 
     * updated as a result of the scan 
     * and configuration passes. 
     */

    }

/*******************************************************************************
*
* sysPcicInit - initialize sh7751 pci controller module as host bridge
*
* This routine initialize the PCI Controller (PCIC) in the SH7751. The
* initialization is done in three steps: Clear CFINIT bit in PCICR
* register, initialize registers, set CFINIT bit in PCICR.
*
* RETURNS: N/A
*
* SEE ALSO: sysMbdHwInit()
*/

void sysPcicInit (void)
    {

    /* Clear CFINIT bit of PCICR if it's set */
 
    if (*PCICR & PCICR_CFINIT)
        {
	*PCICR = PCICR_KEY | (*PCICR & ~PCICR_CFINIT);
        }

#if _BYTE_ORDER == _BIG_ENDIAN
    *PCICR   |= PCICR_KEY | PCICR_BMABT | PCICR_BYTESWAP;
#else
    *PCICR   |= PCICR_KEY | PCICR_BMABT;
#endif

    /* PCI bus control registers for local memory access */

    *PCIBCR1  = PCIC_BCR1;			/* PCI bus control regs */
    *PCIBCR2  = PCIC_BCR2;

    *PCIWCR1  = PCIC_WCR1;			/* PCI wait control regs */
    *PCIWCR2  = PCIC_WCR2;
    *PCIWCR3  = PCIC_WCR3;

    *PCIMCR   = PCIC_MCR;			/* PCI memory control reg. */

    /* Disable all PCIC interrupts and clear the interrupt status bits */

    *PCIINTM  = 0;
    *PCIINT   = 0xffffffff;
    *PCIAINTM = 0;
    *PCIAINT  = 0xffffffff;
    *PCIPINTM = 0;
    *PCIPINT  = 0xffffffff;

    /* PCIC Command Register */
 
    *PCIC_CONFIG_1 |= PCI_CMD_MASTER_ENABLE |	/* Enable Bus Master operation*/
		      PCI_CMD_IO_ENABLE	    |	/* Enable access to I/O space */
		      PCI_CMD_MEM_ENABLE    |	/* Enable access to Mem space */
		      PCI_CMD_WC_ENABLE     |	/* Enable wait cycle */
		      PCI_CMD_SERR_ENABLE   | 	/* Enable system error */
		      PCI_CMD_PERR_ENABLE; 	/* Enable parity error */
 
    /* I/O base address to local registers */

    *PCIC_CONFIG_4 = PCI_MSTR_PCIC_BUS;
 
    /* Slave memory space base address */

    *PCIC_CONFIG_5 = PCI_SLV_MEM_BUS;		/* Memory access in P0 */
    *PCILSR0       = PCI_SLV_MEM_SIZE - 0x00100000;
    *PCILAR0       = PCI_SLV_MEM_LOCAL;

    *PCIC_CONFIG_6 = PCI_SLV_MEM_BUS + SH7751_P2_BASE; /* P2 */
    *PCILSR1       = PCI_SLV_MEM_SIZE - 0x00100000;
    *PCILAR1       = PCI_SLV_MEM_LOCAL;

#ifdef INCLUDE_SH7751_VIRTUAL_PCI_MEMIO
    /*
     * Replace the TLB mishit handler to support virtual PCI memory (and IO)
     * space mapping.  This must be after excVecInit(), and before usrMmuInit().
     * The new TLB mishit handler updates the PCIMBR (or PCIIOBR) upon
     * accessing the extended virtual PCI memory (or IO) space.
     * In this situation, the PCIMBR (and PCIIOBR) setup down below is void
     * but is certainly used by Boot ROMs since they do not enable MMU.
     */

    if (sysPciMapVerify ((void *)PCI_MSTR_MEMIO_LOCAL,
			   (UINT)PCI_MSTR_MEMIO_SIZE) == OK)
	{
#  ifndef INCLUDE_SH7751_VIRTUAL_PCI_IO
	excPciMapInit ((void *)PCI_MSTR_MEMIO_LOCAL,
		       (void *)PCI_MEMIO_ADRS,
			 (UINT)PCI_MSTR_MEMIO_SIZE, NULL, NULL, (UINT)0);
#  else
	if (sysPciMapVerify ((void *)PCI_MSTR_IO_LOCAL,
			       (UINT)PCI_MSTR_IO_SIZE) == OK)
	    {
	    excPciMapInit ((void *)PCI_MSTR_MEMIO_LOCAL,
			   (void *)PCI_MEMIO_ADRS,
			     (UINT)PCI_MSTR_MEMIO_SIZE,
			   (void *)PCI_MSTR_IO_LOCAL,
			   (void *)PCI_MSTR_IO_BUS,
			     (UINT)PCI_MSTR_IO_SIZE);
	    }
#  endif /* INCLUDE_SH7751_VIRTUAL_PCI_IO */
	}
#endif /* INCLUDE_SH7751_VIRTUAL_PCI_MEMIO */

    /* Memory Space Base Register, set upper 8 bits */

    *PCIMBR        = PCI_MEMIO_ADRS;
 
    /* IO Space Base Register, set upper 14 bits */
 
    *PCIIOBR       = PCI_MSTR_IO_BUS;
 
    /* set PCICR_CFINIT to enable the PCIC */

    *PCICR        |= (PCICR_KEY | PCICR_CFINIT);
    }

#ifdef INCLUDE_SH7751_VIRTUAL_PCI_MEMIO
/*******************************************************************************
*
* sysPciMapVerify - check PCI virtual address overlap with sysPhysMemDesc table
*
* This routine checks for any virtual address overlap between a planned PCI
* space mapping and sysPhysMemDesc[] table entries.
*
* RETURNS: OK, or ERROR if overlap.
*/

STATUS sysPciMapVerify
    (
    void *pciVirtLow,		/* base address of virtual PCI space */
    UINT  pciVirtLen		/* byte length of virtual PCI space */
    )
    {
    int i;

    for (i = 0; i < sysPhysMemDescNumEnt; i++)
	{
	if (pciVirtLow + pciVirtLen <= sysPhysMemDesc[i].virtualAddr)
	    continue;

	if (pciVirtLow >= sysPhysMemDesc[i].virtualAddr + sysPhysMemDesc[i].len)
	    continue;

	return ERROR;
	}

    return OK;
    }
#endif /* INCLUDE_SH7751_VIRTUAL_PCI_MEMIO */

/*******************************************************************************
*
* sysPciAutoconfigInclude - PCI autoconfig support routine
*
* RETURNS: OK, or ERROR for excluded devices.
*/

STATUS sysPciAutoconfigInclude
    (
    PCI_SYSTEM * pSys,			/* PCI_SYSTEM structure pointer */
    PCI_LOC * pLoc,			/* pointer to function in question */
    UINT devVend			/* deviceID/vendorID of device */
    )
    {
    if (devVend == ((PIBC_DEVICE_ID << 16) | PIBC_VENDOR_ID) ||
	devVend == ((PMU_DEVICE_ID << 16)  | PMU_VENDOR_ID))
	{
	return (ERROR);
	}

    return OK; /* Autoconfigure all other devices */
    }

/*******************************************************************************
*
* sysPciAutoconfigIntrAssign - PCI autoconfig support routine
*
* RETURNS: PCI interrupt line number given pin mask
*/

UCHAR sysPciAutoconfigIntrAssign
    (
    PCI_SYSTEM * pSys,			/* PCI_SYSTEM structure pointer */
    PCI_LOC * pLoc,			/* pointer to function in question */
    UCHAR pin				/* contents of PCI int pin register */
    )
    {
    UCHAR tmpChar = 0xff;

    /* 
     * Ensure this is a resonable value for bus zero.
     * If OK, return INT level, else we return 0xff.
     */

    if (pLoc->device < 16)
	tmpChar = intLine [(pLoc->device)][(pin - 1)];

    /* return the value to be assigned to the pin */

    return (tmpChar);
    }

/******************************************************************************
*
* sysPcicHwInit2 - connect BSP PCIC  interrupts
*
* This routine connects the PCIC interrupts as Power-On and Power-Down.
* It is called from sysHwInit2().
*
* RETURNS: N/A
*
* SEE ALSO: 
*/

void sysPcicHwInit2 (void)
    {
    intConnect (INUM_TO_IVEC (INUM_PCI_SERR), pcicInt, 0);
    intConnect (INUM_TO_IVEC (INUM_PCI_ERR), pcicInt, 0);

    /* Enable PCI interrupts in INTC */
    
    *INT_PRI_00 = (*INT_PRI_00 & 0xff00) | INT_LVL_PCISERR
					 | INT_LVL_PCIERR << 4;

    *INT_MSKCLR_00 = 0x81; 		/* PCIERR and PCISERR */

    /* Set PCIC interrupt masks */

    *PCIINTM  = 0xc3ff;
    *PCIAINTM = 0x380f;
    }

/******************************************************************************
*
* pcicInt - interrupt handler for PCIC.
*
* This routine handles enabled PCIC interrupts.
*
* RETURNS: N/A
*
* SEE ALSO: sysPcicConnect()
*/

void pcicInt (void)
    {
    UINT32 pciInt    = *PCIINT;
    UINT32 pciArbInt = *PCIAINT;
    UINT32 pciAlr    = *PCIALR;
    UINT32 pciClr    = *PCICLR;
    UINT32 pciBmlr   = *PCIBMLR;

    /* clear interrupts by writing back the 1-s */

    *PCIINT = pciInt;
    *PCIAINT = pciArbInt;

    logMsg("PCIC detected error:\n  PCIINT:  0x%08x\n  PCIAINT: 0x%08x \
            \n  PCIALR:  0x%08x\n  PCICLR:  0x%08x\n  PCIBMLR: 0x%08x\n", 
	    pciInt, pciArbInt, pciAlr, pciClr, pciBmlr, 6);

    }


/*******************************************************************************
*
* sysPciSpecialCycle - generate a special cycle with a message
*
* This routine generates a special cycle with a message.
*
* NOMANUAL
*
* RETURNS: OK
*/

LOCAL STATUS sysPciSpecialCycle
    (
    int		busNo,
    UINT32	message
    )
    {
    int deviceNo	= 0x0000001f;
    int funcNo		= 0x00000007;

    sysPciConfigOutCar (PCIC_CAR,
			pciConfigBdfPack (busNo, deviceNo, funcNo) |
			0x80000000);

    sysPciConfigOutLong (PCIC_CDR, message);

    return (OK);
    }

/*******************************************************************************
*
* sysPciConfigRead - read from the PCI configuration space
*
* This routine reads either a byte, word or a long word specified by
* the argument <width>, from the PCI configuration space
* This routine works around a problem in the hardware which hangs
* PCI bus if device no 12 is accessed from the PCI configuration space.
*
* NOMANUAL
*
* RETURNS: OK, or ERROR if this library is not initialized
*/

LOCAL STATUS sysPciConfigRead
    (
    int	busNo,    /* bus number */
    int	deviceNo, /* device number */
    int	funcNo,	  /* function number */
    int	offset,	  /* offset into the configuration space */
    int width,	  /* width to be read */
    void * pData /* data read from the offset */
    )
    {
    UINT8  retValByte = 0;
    UINT16 retValWord = 0;
    UINT32 retValLong = 0;
    STATUS retStat = ERROR;

    switch (width)
        {
        case 1:	/* byte */
            sysPciConfigOutCar (PCIC_CAR,
                  		pciConfigBdfPack (busNo, deviceNo, funcNo) |
				(offset & 0xfc) | 0x80000000);
            
            retValByte = sysPciConfigInByte (PCIC_CDR + (offset & 0x3));
            *((UINT8 *)pData) = retValByte;
	    retStat = OK;
            break;

        case 2: /* word */
            sysPciConfigOutCar (PCIC_CAR,
				pciConfigBdfPack (busNo, deviceNo, funcNo) |
				(offset & 0xfc) | 0x80000000);

	    retValWord = sysPciConfigInWord (PCIC_CDR + (offset & 0x2));
            *((UINT16 *)pData) = retValWord;
	    retStat = OK;
	    break;

        case 4: /* long */
	    sysPciConfigOutCar (PCIC_CAR,
				pciConfigBdfPack (busNo, deviceNo, funcNo) |
				(offset & 0xfc) | 0x80000000);
	    retValLong = sysPciConfigInLong (PCIC_CDR);
            *((UINT32 *)pData) = retValLong;
	    retStat = OK;
            break;

        default:
            retStat = ERROR;
            break;
        }

    return (retStat);
    }

/*******************************************************************************
*
* sysPciConfigWrite - write to the PCI configuration space
*
* This routine writes either a byte, word or a long word specified by
* the argument <width>, to the PCI configuration space
* This routine works around a problem in the hardware which hangs
* PCI bus if device no 12 is accessed from the PCI configuration space.
*
* NOMANUAL
*
* RETURNS: OK, or ERROR if this library is not initialized
*/

LOCAL STATUS sysPciConfigWrite
    (
    int	busNo,    /* bus number */
    int	deviceNo, /* device number */
    int	funcNo,	  /* function number */
    int	offset,	  /* offset into the configuration space */
    int width,	  /* width to write */
    ULONG data	  /* data to write */
    )
    {
    switch (width)
        {
        case 1:	/* byte */
            sysPciConfigOutCar (PCIC_CAR,
				pciConfigBdfPack (busNo, deviceNo, funcNo) |
				(offset & 0xfc) | 0x80000000);
	    sysPciConfigOutByte ((PCIC_CDR + (offset & 0x3)), data);
            break;

        case 2: /* word */
            sysPciConfigOutCar (PCIC_CAR,
				pciConfigBdfPack (busNo, deviceNo, funcNo) |
				(offset & 0xfc) | 0x80000000);
	    sysPciConfigOutWord ((PCIC_CDR + (offset & 0x2)), data);
	    break;

        case 4: /* long */
	    sysPciConfigOutCar (PCIC_CAR,
				pciConfigBdfPack (busNo, deviceNo, funcNo) |
				(offset & 0xfc) | 0x80000000);
	    sysPciConfigOutLong (PCIC_CDR, data);
            break;

        default:
            return (ERROR);
            
        }
    return (OK);
    }


/*******************************************************************************
*
* sysPciConfigOutByte - write a byte to the passed address
*
* SEE ALSO:
*/
 
LOCAL void sysPciConfigOutByte
    (
    ULONG 		port,		/* port to be written to */
    UCHAR 		data		/* data to be copied */
    )
    {
    volatile UINT32	*tmpDevAdrs;
    UINT32		val;
    UINT32		shift;

    /* 
     * The hardware only allows dword access. We must select the byte 
     * within the dword.
     */

    tmpDevAdrs = (UINT32 *) (port & 0xfffffffc);
    val = *tmpDevAdrs;

#if _BYTE_ORDER == _BIG_ENDIAN
    shift = (~port & 0x3) << 3;
#else
    shift = (port & 0x3) << 3;
#endif
    *tmpDevAdrs = (val & ~(0xff << shift)) | (data << shift);
    }
 
/*******************************************************************************
*
* sysPciConfigOutWord - write a word to the passed address
*
* SEE ALSO:
*/
 
LOCAL void sysPciConfigOutWord
    (
    ULONG 		port,		/* port to be written to */
    UINT16 		data		/* data to be copied */
    )
    {
    volatile UINT32	*tmpDevAdrs;
    UINT32		val;
    UINT32		shift;
 
    /* 
     * The hardware only allows dword access. We must select the word 
     * within the dword.
     */

    tmpDevAdrs = (UINT32 *) (port & 0xfffffffc);
    val = *tmpDevAdrs;


#if   _BYTE_ORDER == _BIG_ENDIAN   
    shift = (~port & 0x2) << 3;
    data  = (data >> 8) | (data << 8);
#else
    shift = (port & 0x2) << 3;
#endif
    *tmpDevAdrs = (val & ~(0xffff << shift)) | (data << shift);
    }
 
 
/*******************************************************************************
*
* sysPciConfigOutLong - write a long to the passed address
*
* SEE ALSO:
*/
 
LOCAL void sysPciConfigOutLong
    (
    ULONG		port,		/* port to be written to */
    ULONG		data		/* data to be copied */
    )
    {
    volatile ULONG * 	tmpDevAdrs;
 
    tmpDevAdrs = (ULONG *) port;

#if _BYTE_ORDER == _BIG_ENDIAN
    *tmpDevAdrs = LONGSWAP(data);
#else
    *tmpDevAdrs = data;
#endif
    }
 
/*******************************************************************************
*
* sysPciConfigInByte - read a byte from the passed address
*
* SEE ALSO:
*/
 
LOCAL UCHAR sysPciConfigInByte
    (
    ULONG		port		/* port to be read from */
    )
    {
    volatile UINT32	*tmpDevAdrs;
    UINT32		val;
    UINT32		shift;
 
    /* 
     * The hardware only allows dword access. We must select the byte 
     * within the dword.
     */

    tmpDevAdrs = (UINT32 *) (port & 0xfffffffc) ;
    val = *tmpDevAdrs;
    
#if _BYTE_ORDER == _BIG_ENDIAN
    shift = (~port & 0x3) << 3;
#else
    shift = (port & 0x3) << 3;
#endif

    return ((val >> shift) & 0xff);
    }
 
/*******************************************************************************
*
* sysPciConfigInWord - read a word from the passed address
*
* SEE ALSO:
*/
 
LOCAL USHORT sysPciConfigInWord
    (
    ULONG		port		/* port to be read from */
    )
    {
    volatile UINT32	*tmpDevAdrs;
    UINT32		val;
    UINT32		shift;
 
    /* 
     * The hardware only allows dword access. We must select the word 
     * within the dword.
     */

    tmpDevAdrs = (UINT32 *) (port & 0xfffffffc) ;
    val = *tmpDevAdrs;
 
#if _BYTE_ORDER == _BIG_ENDIAN
    val = LONGSWAP(val);
#endif

    shift = (port & 0x2) << 3;
    return ((val >> shift) & 0xffff);
    }
 
/*******************************************************************************
*
* sysPciConfigInLong - read a long from the passed address
*
* SEE ALSO:
*/
 
LOCAL ULONG sysPciConfigInLong
    (
    ULONG		port          /* port to be read from */
    )
    {
    volatile ULONG	*tmpDevAdrs;
    ULONG		val;

    tmpDevAdrs = (ULONG *) port;
    val = *tmpDevAdrs;

#if _BYTE_ORDER == _BIG_ENDIAN
    return (LONGSWAP(val));
#else
    return (val);
#endif
    }

/*******************************************************************************
*
* sysPciConfigOutCar - Set the PCI_CAR address
*
* SEE ALSO:
*/
 
LOCAL void sysPciConfigOutCar
    (
    ULONG		adrs,		/* port to be written to */
    ULONG		data		/* data to be copied */
    )
    {
    *(volatile ULONG *)adrs = data;
    }
